part of fields.model;

/// An editor for a [Field].
/// 
/// The type parameter `T` sets the [valueType] for this field editor.
abstract class FieldEditor<T> {
  
  static final _fieldEditorClassMirror = reflectClass(FieldEditor);
  static final _editors = new Map<Type, ClassMirror>();
  static ClassMirror _defaultEditorClassMirror;
    
  static bool isRegistered(Type fieldType) => _editors.containsKey(fieldType);
  static bool get isDefaultRegistered => _defaultEditorClassMirror != null;
  
  static void register(Type fieldType, Type editorType) {
    if (editorType == null) {
      throw new ArgumentError('editorType must be !null');
    }
    if (isRegistered(fieldType)) {
      throw new StateError('$fieldType already has a registered editorType');
    }
    final editorClassMirror = reflectClass(editorType);
    if (!_isAssignableFrom(_fieldEditorClassMirror, editorClassMirror)) {
      throw new ArgumentError(
          '$editorType must extend or implement $FieldEditor');
    }
    _editors[fieldType] = editorClassMirror;
  }
  
  static void registerDefault(Type editorType) {
    if (editorType == null) {
      throw new ArgumentError('editorType must be !null');
    }
    if (isDefaultRegistered) {
      throw new StateError('Default editorType is already registered.');
    }
    final editorClassMirror = reflectClass(editorType);
    if (!_isAssignableFrom(_fieldEditorClassMirror, editorClassMirror)) {
      throw new ArgumentError(
          '$editorType must extend or implement $FieldEditor');
    }
    _defaultEditorClassMirror = editorClassMirror;
  }
  
  static bool unregister(Type fieldType) => _editors.remove(fieldType) != null;
  
  static void unregisterDefault() {
    _defaultEditorClassMirror = null;
  }
  
  static void unregisterAll() {
    unregisterDefault();
    _editors.clear();
  }
  
  /// The element rendered by this editor.
  get element;
  
  /// The [Field] edited by this editor.
  final Field field;
  
  /// The type of [value] edited by this editor.
  /// 
  /// This type may differ from the [Field.type] of this editor's [field].  If
  /// they differ, the [value] must be converted to and from this type by this
  /// editor.
  final Type valueType = T;
  
  bool _isEnabled = true;
  /// Whether or not this editor is enabled.
  /// 
  /// This editor is enabled unless:
  /// 
  /// * This is set to `false`.
  /// * This editor's [field] is read-only.
  /// * This editor's [field] cannot be transformed to this editor's 
  /// [valueType].
  bool 
    get isEnabled => 
        _isEnabled &&
        !field.isReadOnly &&
        ( field.transform != null ||
        ValueTransform.canTransform(valueType, field.type) );        
    set isEnabled(bool isEnabled) {
      if (_isEnabled != isEnabled) {
        _isEnabled = isEnabled;
        refreshElement();
      }
    }

  /// The value edited by this editor.
  /// 
  /// This value transforms the [Field.value] of this editor's [field] to be of
  /// [valueType].
  /// 
  /// Editors should always interact with this value to ensure the proper
  /// transforms are applied.
  dynamic
    get value {
      final value = field.value;
      if (field.transform != null) {
        return field.transform(value);
      } else if (ValueTransform.canTransform(field.type, valueType)) {
        return ValueTransform.transform(value, field.type, valueType);
      } else {
        return value.toString();
      }
    }
    set value(value) {
      if(!isEnabled) {
        throw new UnsupportedError('Cannot set editor value if !isEnabled');
      }
      if (field.transform != null && field.transform.hasInverse) {
        field.value = field.transform.inverse(value);
      } else {
        field.value = ValueTransform.transform(value, valueType, field.type);
      }
    }
  
  /// Creates a new field editor for the given [field].
  /// 
  /// The class of the created field editor is the first of the following that 
  /// matches:
  /// 
  /// * The class given by [Field.editor] if it is not `null`.
  /// * The registered editor class for the [Field.type], if any.
  /// * The registered default editor class, if any.
  /// 
  /// If none of the above conditions match then an error is thrown.
  factory FieldEditor.forField(Field field) {
    ClassMirror cm = _defaultEditorClassMirror;
    if (field.editorClass != null) {
      cm = field.editorClass;
    } else if (isRegistered(field.type)) {
      cm = _editors[field.type];
    }
    if (cm == null) {
      throw new StateError('No editor registered for type <${field.type}>');
    }
    return cm.newInstance(const Symbol(''),[field]).reflectee;
  }
  
  FieldEditor(this.field) {
    field.values.listen((_) => refreshElement());
    scheduleMicrotask(refreshElement);
  }

  /// Refresh this field editor's [element].
  /// 
  /// This method is called whenever this editor's [value] or [isEnabled] state
  /// changes.
  void refreshElement();
}
